from typing import TYPE_CHECKING, Any, Dict, ForwardRef, List, Literal, Tuple, Union

from cattr import Converter

if TYPE_CHECKING:
    from zarrita.codecs import CodecMetadata


ZARR_JSON = "zarr.json"


def is_total_slice(item, shape):
    """Determine whether `item` specifies a complete slice of array with the
    given `shape`. Used to optimize __setitem__ operations on the Chunk
    class."""

    # N.B., assume shape is normalized

    if item == Ellipsis:
        return True
    if item == slice(None):
        return True
    if isinstance(item, slice):
        item = (item,)
    if isinstance(item, tuple):
        return all(
            (
                isinstance(s, slice)
                and (
                    (s == slice(None))
                    or ((s.stop - s.start == l) and (s.step in [1, None]))
                )
            )
            for s, l in zip(item, shape)
        )
    else:
        raise TypeError("expected slice or tuple of slices, found %r" % item)


def get_order(codecs: List["CodecMetadata"]) -> Literal["C", "F"]:
    for codec in codecs:
        if codec.name == "transpose":
            order = codec.configuration.order
            if not isinstance(order, tuple):
                return order
    return "C"


def make_cattr():
    from zarrita.array import (
        ChunkKeyEncodingMetadata,
        DefaultChunkKeyEncodingMetadata,
        V2ChunkKeyEncodingMetadata,
    )
    from zarrita.codecs import (
        BloscCodecMetadata,
        CodecMetadata,
        EndianCodecMetadata,
        GzipCodecMetadata,
        ShardingCodecMetadata,
        TransposeCodecMetadata,
    )

    dataset_converter = Converter()

    def _structure_chunk_key_encoding_metadata(
        d: Dict[str, Any], _t
    ) -> ChunkKeyEncodingMetadata:
        if d["name"] == "default":
            return dataset_converter.structure(d, DefaultChunkKeyEncodingMetadata)
        if d["name"] == "v2":
            return dataset_converter.structure(d, V2ChunkKeyEncodingMetadata)
        raise KeyError

    dataset_converter.register_structure_hook(
        ChunkKeyEncodingMetadata, _structure_chunk_key_encoding_metadata
    )

    def _structure_codec_metadata(d: Dict[str, Any], _t=None) -> CodecMetadata:
        if d["name"] == "blosc":
            return dataset_converter.structure(d, BloscCodecMetadata)
        if d["name"] == "endian":
            return dataset_converter.structure(d, EndianCodecMetadata)
        if d["name"] == "transpose":
            return dataset_converter.structure(d, TransposeCodecMetadata)
        if d["name"] == "gzip":
            return dataset_converter.structure(d, GzipCodecMetadata)
        if d["name"] == "sharding_indexed":
            return dataset_converter.structure(d, ShardingCodecMetadata)
        raise KeyError

    dataset_converter.register_structure_hook(CodecMetadata, _structure_codec_metadata)

    dataset_converter.register_structure_hook_factory(
        lambda t: str(t) == "ForwardRef('CodecMetadata')",
        lambda t: _structure_codec_metadata,
    )

    def _structure_order(d: Any, _t=None) -> Union[Literal["C", "F"], Tuple[int, ...]]:
        if d == "C":
            return "C"
        if d == "F":
            return "F"
        if isinstance(d, list):
            return tuple(d)
        raise KeyError

    dataset_converter.register_structure_hook_factory(
        lambda t: str(t)
        == "typing.Union[typing.Literal['C', 'F'], typing.Tuple[int, ...]]",
        lambda t: _structure_order,
    )

    return dataset_converter



from typing import List, Optional, Tuple

import fsspec
import numpy as np


class Store:
    def multi_get(
        self, keys: List[Tuple[str, Optional[Tuple[int, int]]]]
    ) -> List[Optional[bytes]]:
        return [self.get(key, byte_range) for key, byte_range in keys]

    def get(
        self, key: str, byte_range: Optional[Tuple[int, int]] = None
    ) -> Optional[bytes]:
        raise NotImplementedError

    def multi_set(
        self, key_values: List[Tuple[str, bytes, Optional[Tuple[int, int]]]]
    ) -> None:
        for key, value, byte_range in key_values:
            self.set(key, value, byte_range)

    def set(
        self, key: str, value: bytes, byte_range: Optional[Tuple[int, int]] = None
    ) -> None:
        raise NotImplemented

    def delete(self, key: str) -> None:
        raise NotImplemented


class FileSystemStore(Store):
    def __init__(self, url: str, **storage_options):
        assert isinstance(url, str)

        # instantiate file system
        fs, root = fsspec.core.url_to_fs(url, **storage_options)
        self.fs = fs
        self.root = root.rstrip("/")

    def multi_get(
        self, keys: List[Tuple[str, Optional[Tuple[int, int]]]]
    ) -> List[Optional[bytes]]:
        return [self.get(key, byte_range) for key, byte_range in keys]

    def get(
        self, key: str, byte_range: Optional[Tuple[int, int]] = None
    ) -> Optional[bytes]:
        assert isinstance(key, str)
        path = f"{self.root}/{key}"

        try:
            value = (
                self.fs.cat_file(path, byte_range[0], byte_range[1])
                if byte_range
                else self.fs.cat_file(path)
            )
        except (FileNotFoundError, IsADirectoryError, NotADirectoryError):
            return None

        return value

    def multi_set(
        self, key_values: List[Tuple[str, bytes, Optional[Tuple[int, int]]]]
    ) -> None:
        for key, value, byte_range in key_values:
            self.set(key, value, byte_range)

    def set(
        self, key: str, value: bytes, byte_range: Optional[Tuple[int, int]] = None
    ) -> None:
        assert isinstance(key, str)
        path = f"{self.root}/{key}"

        # ensure parent folder exists
        self.fs.mkdirs(self.fs._parent(path), exist_ok=True)

        # write data
        if byte_range:
            with self.fs.open(path, "r+b") as f:
                f.seek(byte_range[0])
                f.write(value)
        else:
            with self.fs.open(path, "wb") as f:
                f.write(value)

    def delete(self, key: str) -> None:
        path = f"{self.root}/{key}"
        if self.fs.exists(path):
            self.fs.rm(path)



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Tuple, Union

import numpy as np
from attr import asdict, frozen
from cattrs import register_structure_hook, structure
from numcodecs.blosc import Blosc
from numcodecs.gzip import GZip

from zarrita.sharding import ShardingCodecConfigurationMetadata, ShardingCodecMetadata
from zarrita.value_handle import ArrayHandle, BufferHandle, NoneHandle, ValueHandle

if TYPE_CHECKING:
    from zarrita.array import CoreArrayMetadata


def _needs_bytes(
    f: Callable[[Any, bytes, Tuple[slice, ...], "CoreArrayMetadata"], ValueHandle]
) -> Callable[[Any, ValueHandle, Tuple[slice, ...], "CoreArrayMetadata"], ValueHandle]:
    def inner(
        _self,
        value: ValueHandle,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        buf = value.tobytes()
        if buf is None:
            return NoneHandle()
        return f(_self, buf, selection, array_metadata)

    return inner


def _needs_array(
    f: Callable[[Any, np.ndarray, Tuple[slice, ...], "CoreArrayMetadata"], ValueHandle]
) -> Callable[[Any, ValueHandle, Tuple[slice, ...], "CoreArrayMetadata"], ValueHandle]:
    def inner(
        _self,
        value: ValueHandle,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        array = value.toarray()
        if array is None:
            return NoneHandle()
        return f(_self, array, selection, array_metadata)

    return inner


@frozen
class BloscCodecConfigurationMetadata:
    cname: Literal["lz4", "lz4hc", "blosclz", "zstd", "snappy", "zlib"] = "zstd"
    clevel: int = 5
    shuffle: Literal[0, 1, 2, -1] = 0
    blocksize: int = 0


@frozen
class BloscCodecMetadata:
    configuration: BloscCodecConfigurationMetadata
    name: Literal["blosc"] = "blosc"

    @_needs_bytes
    def decode(
        self,
        buf: bytes,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return BufferHandle(Blosc.from_config(asdict(self.configuration)).decode(buf))

    @_needs_array
    def encode(
        self,
        chunk: np.ndarray,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        if not chunk.flags.c_contiguous and not chunk.flags.f_contiguous:
            chunk = chunk.copy(order="K")
        return BufferHandle(Blosc.from_config(asdict(self.configuration)).encode(chunk))


@frozen
class EndianCodecConfigurationMetadata:
    endian: Literal["big", "little"] = "little"


@frozen
class EndianCodecMetadata:
    configuration: EndianCodecConfigurationMetadata
    name: Literal["endian"] = "endian"

    def _get_byteorder(self, array: np.ndarray) -> Literal["big", "little"]:
        if array.dtype.byteorder == "<":
            return "little"
        elif array.dtype.byteorder == ">":
            return "big"
        else:
            import sys

            return sys.byteorder

    @_needs_array
    def decode(
        self,
        chunk: np.ndarray,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        byteorder = self._get_byteorder(chunk)
        if self.configuration.endian != byteorder:
            chunk = chunk.view(dtype=chunk.dtype.newbyteorder(byteorder))
        return ArrayHandle(chunk)

    @_needs_array
    def encode(
        self,
        chunk: np.ndarray,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        byteorder = self._get_byteorder(chunk)
        if self.configuration.endian != byteorder:
            chunk = chunk.view(dtype=chunk.dtype.newbyteorder(byteorder))
        return ArrayHandle(chunk)


@frozen
class TransposeCodecConfigurationMetadata:
    order: Union[Literal["C", "F"], Tuple[int, ...]] = "C"


@frozen
class TransposeCodecMetadata:
    configuration: TransposeCodecConfigurationMetadata
    name: Literal["transpose"] = "transpose"

    @_needs_array
    def decode(
        self,
        chunk: np.ndarray,
        _selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        new_order = self.configuration.order
        chunk = chunk.view(np.dtype(array_metadata.data_type.value))
        if isinstance(new_order, tuple):
            chunk = chunk.transpose(new_order)
        else:
            chunk = chunk.reshape(
                array_metadata.chunk_shape,
                order=new_order,
            )
        return ArrayHandle(chunk)

    @_needs_array
    def encode(
        self,
        chunk: np.ndarray,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        new_order = self.configuration.order
        if isinstance(new_order, tuple):
            chunk = chunk.reshape(-1, order="C")
        else:
            chunk = chunk.reshape(-1, order=new_order)
        return ArrayHandle(chunk)


@frozen
class GzipCodecConfigurationMetadata:
    level: int = 5


@frozen
class GzipCodecMetadata:
    configuration: GzipCodecConfigurationMetadata
    name: Literal["gzip"] = "gzip"

    @_needs_bytes
    def decode(
        self,
        buf: bytes,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return BufferHandle(GZip(self.configuration.level).decode(buf))

    @_needs_bytes
    def encode(
        self,
        buf: bytes,
        _selection: Tuple[slice, ...],
        _array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return BufferHandle(GZip(self.configuration.level).encode(buf))


CodecMetadata = Union[
    BloscCodecMetadata,
    EndianCodecMetadata,
    TransposeCodecMetadata,
    GzipCodecMetadata,
    ShardingCodecMetadata,
]


def blosc_codec(
    cname: Literal["lz4", "lz4hc", "blosclz", "zstd", "snappy", "zlib"] = "zstd",
    clevel: int = 5,
    shuffle: Literal[0, 1, 2, -1] = 0,
    blocksize: int = 0,
) -> BloscCodecMetadata:
    return BloscCodecMetadata(
        configuration=BloscCodecConfigurationMetadata(
            cname=cname, clevel=clevel, shuffle=shuffle, blocksize=blocksize
        )
    )


def endian_codec(endian: Literal["big", "little"]) -> EndianCodecMetadata:
    return EndianCodecMetadata(configuration=EndianCodecConfigurationMetadata(endian))


def transpose_codec(
    order: Union[Tuple[int, ...], Literal["C", "F"]]
) -> TransposeCodecMetadata:
    return TransposeCodecMetadata(
        configuration=TransposeCodecConfigurationMetadata(order)
    )


def gzip_codec(level: int = 5) -> GzipCodecMetadata:
    return GzipCodecMetadata(configuration=GzipCodecConfigurationMetadata(level))


def sharding_codec(
    chunk_shape: Tuple[int, ...], codecs: List[CodecMetadata] = []
) -> ShardingCodecMetadata:
    return ShardingCodecMetadata(
        configuration=ShardingCodecConfigurationMetadata(chunk_shape, codecs)
    )



import functools
import itertools
import math
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Literal,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

import numpy as np
from attrs import field, frozen

from zarrita.common import get_order, is_total_slice
from zarrita.indexing import BasicIndexer
from zarrita.value_handle import ArrayHandle, BufferHandle, NoneHandle, ValueHandle

if TYPE_CHECKING:
    from zarrita.array import CoreArrayMetadata
    from zarrita.codecs import CodecMetadata


MAX_UINT_64 = 2**64 - 1


@frozen
class ShardingCodecConfigurationMetadata:
    chunk_shape: Tuple[int, ...]
    codecs: List["CodecMetadata"] = field(factory=list)


def product(tup: Tuple[int, ...]) -> int:
    return functools.reduce(lambda x, y: x * y, tup, 1)


def c_order_iter(chunk_shape: Tuple[int, ...]) -> Iterator[Tuple[int, ...]]:
    return itertools.product(*(range(x) for x in chunk_shape))


def morton_order_iter(chunk_shape: Tuple[int, ...]) -> Iterator[Tuple[int, ...]]:
    def decode_morton(z: int, chunk_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        # Inspired by compressed morton code as implemented in Neuroglancer
        # https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/volume.md#compressed-morton-code
        bits = tuple(math.ceil(math.log2(c)) for c in chunk_shape)
        max_coords_bits = max(*bits)
        input_bit = 0
        input_value = z
        out = [0 for _ in range(len(chunk_shape))]

        for coord_bit in range(max_coords_bits):
            for dim in range(len(chunk_shape)):
                if coord_bit < bits[dim]:
                    bit = (input_value >> input_bit) & 1
                    out[dim] |= bit << coord_bit
                    input_bit += 1
        return tuple(out)

    for i in range(product(chunk_shape)):
        yield decode_morton(i, chunk_shape)


class _ShardIndex(NamedTuple):
    # dtype uint64, shape (chunks_per_shard_0, chunks_per_shard_1, ..., 2)
    offsets_and_lengths: np.ndarray

    def _localize_chunk(self, chunk: Tuple[int, ...]) -> Tuple[int, ...]:
        return tuple(
            chunk_i % shard_i
            for chunk_i, shard_i in zip(chunk, self.offsets_and_lengths.shape)
        )

    def is_all_empty(self) -> bool:
        return np.array_equiv(self.offsets_and_lengths, MAX_UINT_64)

    def get_chunk_slice(self, chunk: Tuple[int, ...]) -> Optional[slice]:
        localized_chunk = self._localize_chunk(chunk)
        chunk_start, chunk_len = self.offsets_and_lengths[localized_chunk]
        if (chunk_start, chunk_len) == (MAX_UINT_64, MAX_UINT_64):
            return None
        else:
            return slice(int(chunk_start), int(chunk_start + chunk_len))

    def set_chunk_slice(
        self, chunk: Tuple[int, ...], chunk_slice: Optional[slice]
    ) -> None:
        localized_chunk = self._localize_chunk(chunk)
        if chunk_slice is None:
            self.offsets_and_lengths[localized_chunk] = (MAX_UINT_64, MAX_UINT_64)
        else:
            self.offsets_and_lengths[localized_chunk] = (
                chunk_slice.start,
                chunk_slice.stop - chunk_slice.start,
            )

    def to_bytes(self) -> bytes:
        return self.offsets_and_lengths.tobytes(order="C")

    @property
    def byte_length(self) -> int:
        return self.index_byte_length(self.offsets_and_lengths.shape)

    @classmethod
    def from_bytes(
        cls,
        buffer: Union[bytes, bytearray],
        chunks_per_shard: Tuple[int, ...],
    ) -> "_ShardIndex":
        try:
            return cls(
                offsets_and_lengths=np.frombuffer(
                    bytearray(buffer), dtype="<u8"
                ).reshape(*chunks_per_shard, 2, order="C"),
            )
        except ValueError as e:  # pragma: no cover
            raise RuntimeError from e

    @classmethod
    def create_empty(cls, chunks_per_shard: Tuple[int, ...]) -> "_ShardIndex":
        # reserving 2*64bit per chunk for offset and length:
        return cls.from_bytes(
            MAX_UINT_64.to_bytes(8, byteorder="little")
            * (2 * product(chunks_per_shard)),
            chunks_per_shard=chunks_per_shard,
        )

    @classmethod
    def index_byte_length(cls, chunks_per_shard: Tuple[int, ...]) -> int:
        return 16 * product(chunks_per_shard)


@frozen
class ShardingCodecMetadata:
    configuration: ShardingCodecConfigurationMetadata
    name: Literal["sharding_indexed"] = "sharding_indexed"

    def decode(
        self,
        value_handle: ValueHandle,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        if isinstance(value_handle, NoneHandle):
            return NoneHandle()

        shard_shape = array_metadata.chunk_shape
        chunk_shape = self.configuration.chunk_shape
        chunks_per_shard = tuple(s // c for s, c in zip(shard_shape, chunk_shape))

        indexer = BasicIndexer(
            selection,
            shape=shard_shape,
            chunk_shape=chunk_shape,
        )

        # setup output array
        out = np.zeros(
            shard_shape,
            dtype=array_metadata.dtype,
            order=get_order(self.configuration.codecs),
        )

        indexed_chunks = list(indexer)
        all_chunk_coords = set(chunk_coords for chunk_coords, _, _ in indexed_chunks)

        # reading bytes of all requested chunks
        shard_dict: Dict[Tuple[int, ...], ValueHandle] = {}
        if self._is_total_shard(all_chunk_coords, chunks_per_shard):
            # read entire shard
            shard_dict = self._load_full_shard(value_handle, chunks_per_shard)
        else:
            # read some chunks within the shard
            shard_index = self._load_shard_index(value_handle, chunks_per_shard)
            shard_dict = {}
            for chunk_coords in all_chunk_coords:
                chunk_byte_slice = shard_index.get_chunk_slice(chunk_coords)
                if chunk_byte_slice:
                    chunk_bytes = value_handle[chunk_byte_slice].tobytes()
                    if chunk_bytes:
                        shard_dict[chunk_coords] = BufferHandle(chunk_bytes)

        # decoding chunks and writing them into the output buffer
        for chunk_coords, chunk_selection, out_selection in indexed_chunks:
            chunk_value = shard_dict.get(chunk_coords, NoneHandle())
            chunk = self._decode_chunk(chunk_value, chunk_selection, array_metadata)
            if chunk is not None:
                tmp = chunk[chunk_selection]
                out[out_selection] = tmp
            elif array_metadata.fill_value is not None:
                out[out_selection] = array_metadata.fill_value

        return ArrayHandle(out)

    def _decode_chunk(
        self,
        value_handle: ValueHandle,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> Optional[np.ndarray]:
        if isinstance(value_handle, NoneHandle):
            return None

        from zarrita.array import CoreArrayMetadata

        # rewriting the metadata to scope it to the shard
        core_metadata = CoreArrayMetadata(
            shape=array_metadata.chunk_shape,
            chunk_shape=self.configuration.chunk_shape,
            data_type=array_metadata.data_type,
            fill_value=array_metadata.fill_value,
        )

        # applying codecs in reverse order
        for codec_metadata in self.configuration.codecs[::-1]:
            value_handle = codec_metadata.decode(value_handle, selection, core_metadata)

        chunk = value_handle.toarray()
        if chunk is None:
            return None

        # ensure correct dtype
        if str(chunk.dtype) != array_metadata.data_type.name:
            chunk = chunk.view(np.dtype(array_metadata.data_type.name))

        # ensure correct chunk shape
        if chunk.shape != self.configuration.chunk_shape:
            chunk = chunk.reshape(self.configuration.chunk_shape)

        return chunk

    def encode(
        self,
        value: ValueHandle,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        shard = value.toarray()
        if shard is None:
            return NoneHandle()

        shard_shape = array_metadata.chunk_shape
        chunk_shape = self.configuration.chunk_shape
        chunks_per_shard = tuple(s // c for s, c in zip(shard_shape, chunk_shape))

        indexer = list(
            BasicIndexer(
                selection,
                shape=shard_shape,
                chunk_shape=chunk_shape,
            )
        )

        assert self._is_total_shard(
            set(chunk_coords for chunk_coords, _, _ in indexer), chunks_per_shard
        )

        # assembling and encoding chunks within the shard
        shard_dict: Dict[Tuple[int, ...], ValueHandle] = {}
        for chunk_coords, chunk_selection, out_selection in indexer:
            if is_total_slice(chunk_selection, chunk_shape):
                chunk = shard[out_selection]
            else:
                # handling writing partial chunks
                chunk = np.empty(
                    chunk_shape,
                    dtype=array_metadata.dtype,
                    order="C",
                )
                if array_metadata.fill_value:
                    chunk.fill(array_metadata.fill_value)
                chunk[chunk_selection] = shard[out_selection]
            if array_metadata.fill_value and np.all(chunk == array_metadata.fill_value):
                shard_dict[chunk_coords] = NoneHandle()
            else:
                shard_dict[chunk_coords] = self._encode_chunk(
                    chunk, chunk_selection, array_metadata
                )

        if all(
            isinstance(chunk_value, NoneHandle) for chunk_value in shard_dict.values()
        ):
            return NoneHandle()
        return BufferHandle(self._build_shard(shard_dict, chunks_per_shard))

    def _encode_chunk(
        self,
        chunk: np.ndarray,
        selection: Tuple[slice, ...],
        array_metadata: "CoreArrayMetadata",
    ):
        from zarrita.array import CoreArrayMetadata

        # rewriting the metadata to scope it to the shard
        core_metadata = CoreArrayMetadata(
            shape=array_metadata.chunk_shape,
            chunk_shape=self.configuration.chunk_shape,
            data_type=array_metadata.data_type,
            fill_value=array_metadata.fill_value,
        )

        encoded_chunk_value: ValueHandle = ArrayHandle(chunk)
        for codec in self.configuration.codecs:
            encoded_chunk_value = codec.encode(
                encoded_chunk_value,
                selection,
                core_metadata,
            )

        return encoded_chunk_value

    def _is_total_shard(
        self, all_chunk_coords: Set[Tuple[int, ...]], chunks_per_shard: Tuple[int, ...]
    ) -> bool:
        return len(all_chunk_coords) == product(chunks_per_shard) and all(
            chunk_coords in all_chunk_coords
            for chunk_coords in c_order_iter(chunks_per_shard)
        )

    def _load_shard_index(
        self, value_handle: ValueHandle, chunks_per_shard: Tuple[int, ...]
    ) -> _ShardIndex:
        index_bytes = value_handle[-_ShardIndex.index_byte_length(chunks_per_shard) :]
        assert isinstance(index_bytes, bytes)
        if index_bytes is not None:
            return _ShardIndex.from_bytes(index_bytes, chunks_per_shard)
        else:
            return _ShardIndex.create_empty(chunks_per_shard)

    def _build_shard(
        self,
        shard_dict: Dict[Tuple[int, ...], ValueHandle],
        chunks_per_shard: Tuple[int, ...],
    ) -> bytes:
        shard_index = _ShardIndex.create_empty(chunks_per_shard)
        byte_shard_dict = {
            chunk_coords: chunk_value.tobytes()
            for chunk_coords, chunk_value in shard_dict.items()
        }

        # output buffer
        shard_bytes = bytearray(
            sum(
                len(chunk_bytes)
                for chunk_bytes in byte_shard_dict.values()
                if chunk_bytes is not None
            )
            + shard_index.byte_length
        )

        # write chunks within shard in morton order
        byte_offset = 0
        for chunk_coords in morton_order_iter(chunks_per_shard):
            chunk_bytes = byte_shard_dict.get(chunk_coords, None)
            if chunk_bytes is not None:
                byte_offset_end = byte_offset + len(chunk_bytes)
                shard_bytes[byte_offset : byte_offset + len(chunk_bytes)] = chunk_bytes
                shard_index.set_chunk_slice(
                    chunk_coords, slice(byte_offset, byte_offset_end)
                )
                byte_offset = byte_offset_end
        shard_bytes[-shard_index.byte_length :] = shard_index.to_bytes()
        return bytes(shard_bytes)

    def _load_full_shard(
        self, value_handle: ValueHandle, chunks_per_shard: Tuple[int, ...]
    ) -> Dict[Tuple[int, ...], ValueHandle]:
        shard_bytes = value_handle.tobytes()
        if shard_bytes:
            assert isinstance(shard_bytes, bytes)
            index_bytes = shard_bytes[
                -_ShardIndex.index_byte_length(chunks_per_shard) :
            ]
            shard_index = _ShardIndex.from_bytes(index_bytes, chunks_per_shard)

            shard_dict: Dict[Tuple[int, ...], ValueHandle] = {}
            for chunk_coords in c_order_iter(chunks_per_shard):
                chunk_byte_slice = shard_index.get_chunk_slice(chunk_coords)
                if chunk_byte_slice is not None:
                    shard_dict[chunk_coords] = BufferHandle(
                        shard_bytes[chunk_byte_slice]
                    )

            return shard_dict
        else:
            return {}



import json
from enum import Enum
from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

import numpy as np
from attr import asdict, field, frozen

from zarrita.codecs import CodecMetadata
from zarrita.common import ZARR_JSON, get_order, is_total_slice, make_cattr
from zarrita.indexing import BasicIndexer
from zarrita.store import Store
from zarrita.value_handle import ArrayHandle, FileHandle, NoneHandle, ValueHandle


class DataType(Enum):
    bool = "bool"
    int8 = "int8"
    int16 = "int16"
    int32 = "int32"
    int64 = "int64"
    uint8 = "uint8"
    uint16 = "uint16"
    uint32 = "uint32"
    uint64 = "uint64"
    float32 = "float32"
    float64 = "float64"


dtype_to_data_type = {
    "bool": "bool",
    "|i1": "int8",
    "<i2": "int16",
    "<i4": "int32",
    "<i8": "int64",
    "|u1": "uint8",
    "<u2": "uint16",
    "<u4": "uint32",
    "<u8": "uint64",
    "<f4": "float32",
    "<f8": "float64",
}


@frozen
class RegularChunkGridConfigurationMetadata:
    chunk_shape: Tuple[int, ...]


@frozen
class RegularChunkGridMetadata:
    configuration: RegularChunkGridConfigurationMetadata
    name: Literal["regular"] = "regular"


@frozen
class DefaultChunkKeyEncodingConfigurationMetadata:
    separator: Literal[".", "/"] = "/"


@frozen
class DefaultChunkKeyEncodingMetadata:
    configuration: DefaultChunkKeyEncodingConfigurationMetadata = (
        DefaultChunkKeyEncodingConfigurationMetadata()
    )
    name: Literal["default"] = "default"

    def decode_chunk_key(self, chunk_key: str) -> Tuple[int, ...]:
        if chunk_key == "c":
            return ()
        return tuple(map(int, chunk_key[1:].split(self.configuration.separator)))

    def encode_chunk_key(self, chunk_coords: Tuple[int, ...]) -> str:
        return self.configuration.separator.join(map(str, ("c",) + chunk_coords))


@frozen
class V2ChunkKeyEncodingConfigurationMetadata:
    separator: Literal[".", "/"] = "."


@frozen
class V2ChunkKeyEncodingMetadata:
    configuration: V2ChunkKeyEncodingConfigurationMetadata = (
        V2ChunkKeyEncodingConfigurationMetadata()
    )
    name: Literal["v2"] = "v2"

    def decode_chunk_key(self, chunk_key: str) -> Tuple[int, ...]:
        return tuple(map(int, chunk_key.split(self.configuration.separator)))

    def encode_chunk_key(self, chunk_coords: Tuple[int, ...]) -> str:
        chunk_identifier = self.configuration.separator.join(map(str, chunk_coords))
        return "0" if chunk_identifier == "" else chunk_identifier


ChunkKeyEncodingMetadata = Union[
    DefaultChunkKeyEncodingMetadata, V2ChunkKeyEncodingMetadata
]


@frozen
class CoreArrayMetadata:
    shape: Tuple[int, ...]
    chunk_shape: Tuple[int, ...]
    data_type: DataType
    fill_value: Any

    @property
    def dtype(self) -> np.dtype:
        return np.dtype(self.data_type.value)


@frozen
class ArrayMetadata:
    shape: Tuple[int, ...]
    data_type: DataType
    chunk_grid: RegularChunkGridMetadata
    chunk_key_encoding: ChunkKeyEncodingMetadata
    fill_value: Any
    attributes: Dict[str, Any] = field(factory=dict)
    codecs: List[CodecMetadata] = field(factory=list)
    dimension_names: Optional[Tuple[str, ...]] = None
    zarr_format: Literal[3] = 3
    node_type: Literal["array"] = "array"

    @property
    def dtype(self) -> np.dtype:
        return np.dtype(self.data_type.value)


class Array:
    metadata: ArrayMetadata
    store: "Store"
    path: str

    @classmethod
    def create(
        cls,
        store: "Store",
        path: str,
        *,
        shape: Tuple[int, ...],
        dtype: Union[str, np.dtype],
        chunk_shape: Tuple[int, ...],
        fill_value: Optional[Any] = None,
        chunk_key_encoding: Union[
            Tuple[Literal["default"], Literal[".", "/"]],
            Tuple[Literal["v2"], Literal[".", "/"]],
        ] = ("default", "/"),
        codecs: Optional[Iterable[CodecMetadata]] = None,
        dimension_names: Optional[Iterable[str]] = None,
        attributes: Optional[Dict[str, Any]] = None,
    ) -> "Array":
        data_type = (
            DataType[dtype]
            if isinstance(dtype, str)
            else DataType[dtype_to_data_type[dtype.str]]
        )

        metadata = ArrayMetadata(
            shape=shape,
            data_type=data_type,
            chunk_grid=RegularChunkGridMetadata(
                configuration=RegularChunkGridConfigurationMetadata(
                    chunk_shape=chunk_shape
                )
            ),
            chunk_key_encoding=(
                V2ChunkKeyEncodingMetadata(
                    configuration=V2ChunkKeyEncodingConfigurationMetadata(
                        separator=chunk_key_encoding[1]
                    )
                )
                if chunk_key_encoding[0] == "v2"
                else DefaultChunkKeyEncodingMetadata(
                    configuration=DefaultChunkKeyEncodingConfigurationMetadata(
                        separator=chunk_key_encoding[1]
                    )
                )
            ),
            fill_value=fill_value,
            codecs=list(codecs) if codecs else [],
            dimension_names=tuple(dimension_names) if dimension_names else None,
            attributes=attributes or {},
        )
        array = cls()
        array.metadata = metadata
        array.store = store
        array.path = path
        array._save_metadata()
        return array

    @classmethod
    def open(cls, store: "Store", path: str) -> "Array":
        zarr_json_bytes = store.get(f"{path}/{ZARR_JSON}")
        assert zarr_json_bytes is not None
        return cls.from_json(store, path, json.loads(zarr_json_bytes))

    @classmethod
    def from_json(cls, store: Store, path: str, zarr_json: Any) -> "Array":
        array = cls()
        array.metadata = make_cattr().structure(zarr_json, ArrayMetadata)
        array.store = store
        array.path = path
        return array

    def _save_metadata(self) -> None:
        def convert(o):
            if isinstance(o, DataType):
                return o.name
            raise TypeError

        self.store.set(
            f"{self.path}/{ZARR_JSON}",
            json.dumps(asdict(self.metadata), default=convert).encode(),
        )

    @property
    def ndim(self) -> int:
        return len(self.metadata.shape)

    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]):
        indexer = BasicIndexer(
            selection,
            shape=self.metadata.shape,
            chunk_shape=self.metadata.chunk_grid.configuration.chunk_shape,
        )

        # setup output array
        out = np.zeros(
            indexer.shape,
            dtype=self.metadata.dtype,
            order=get_order(self.metadata.codecs),
        )

        # reading chunks and decoding them
        for chunk_coords, chunk_selection, out_selection in indexer:
            chunk_key = f"{self.path}/{self.metadata.chunk_key_encoding.encode_chunk_key(chunk_coords)}"
            value_handle = FileHandle(self.store, chunk_key)
            chunk = self._decode_chunk(value_handle, chunk_selection)
            if chunk is not None:
                tmp = chunk[chunk_selection]
                out[out_selection] = tmp
            elif self.metadata.fill_value is not None:
                out[out_selection] = self.metadata.fill_value

        if out.shape:
            return out
        else:
            return out[()]

    def _decode_chunk(
        self, value_handle: ValueHandle, selection: Tuple[slice, ...]
    ) -> Optional[np.ndarray]:
        if isinstance(value_handle, NoneHandle):
            return None

        core_metadata = CoreArrayMetadata(
            shape=self.metadata.shape,
            chunk_shape=self.metadata.chunk_grid.configuration.chunk_shape,
            data_type=self.metadata.data_type,
            fill_value=self.metadata.fill_value,
        )

        # apply codecs in reverse order
        for codec_metadata in self.metadata.codecs[::-1]:
            value_handle = codec_metadata.decode(value_handle, selection, core_metadata)

        chunk = value_handle.toarray()
        if chunk is None:
            return None

        # ensure correct dtype
        if str(chunk.dtype) != self.metadata.data_type.name:
            chunk = chunk.view(self.metadata.dtype)

        # ensure correct chunk shape
        if chunk.shape != self.metadata.chunk_grid.configuration.chunk_shape:
            chunk = chunk.reshape(self.metadata.chunk_grid.configuration.chunk_shape)

        return chunk

    def __setitem__(
        self, selection: Union[slice, Tuple[slice, ...]], value: np.ndarray
    ) -> None:
        chunk_shape = self.metadata.chunk_grid.configuration.chunk_shape
        indexer = BasicIndexer(
            selection,
            shape=self.metadata.shape,
            chunk_shape=chunk_shape,
        )

        sel_shape = indexer.shape

        # check value shape
        if sel_shape == ():
            # setting a single item
            assert np.isscalar(value)
        elif np.isscalar(value):
            # setting a scalar value
            pass
        else:
            if not hasattr(value, "shape"):
                value = np.asarray(value, self.metadata.dtype)
            assert value.shape == sel_shape

        # merging with existing data and encoding chunks
        for chunk_coords, chunk_selection, out_selection in indexer:
            chunk_key = f"{self.path}/{self.metadata.chunk_key_encoding.encode_chunk_key(chunk_coords)}"
            value_handle = FileHandle(self.store, chunk_key)

            if is_total_slice(chunk_selection, chunk_shape):
                # write entire chunks
                if sel_shape == ():
                    chunk = value
                elif np.isscalar(value):
                    chunk = np.empty(
                        chunk_shape,
                        dtype=self.metadata.dtype,
                        order="C",
                    )
                    chunk.fill(value)
                else:
                    chunk = value[out_selection].astype(
                        self.metadata.dtype, order="C", copy=False
                    )
            else:
                # writing partial chunks
                # read chunk first
                tmp = self._decode_chunk(
                    value_handle,
                    tuple(slice(0, c) for c in chunk_shape),
                )

                # merge new value
                if tmp is None:
                    chunk = np.empty(
                        chunk_shape,
                        dtype=self.metadata.dtype,
                        order="C",
                    )
                    if self.metadata.fill_value:
                        chunk.fill(self.metadata.fill_value)
                else:
                    chunk = tmp.copy(order="K")  # make a writable copy
                chunk[chunk_selection] = value[out_selection]

            chunk_value: ValueHandle
            if self.metadata.fill_value and np.all(chunk == self.metadata.fill_value):
                # chunks that only contain fill_value will be removed
                chunk_value = NoneHandle()
            else:
                chunk_value = self._encode_chunk(
                    chunk,
                    tuple(slice(0, c) for c in chunk_shape),
                )

            # write out chunk
            value_handle[:] = chunk_value

    def _encode_chunk(self, chunk_value: np.ndarray, selection: Tuple[slice, ...]):
        core_metadata = CoreArrayMetadata(
            shape=self.metadata.shape,
            chunk_shape=self.metadata.chunk_grid.configuration.chunk_shape,
            data_type=self.metadata.data_type,
            fill_value=self.metadata.fill_value,
        )
        encoded_chunk_value: ValueHandle = ArrayHandle(chunk_value)
        for codec in self.metadata.codecs:
            encoded_chunk_value = codec.encode(
                encoded_chunk_value,
                selection,
                core_metadata,
            )

        return encoded_chunk_value

    def __repr__(self):
        path = self.path
        return f"<Array {path}>"



from abc import abstractmethod
from typing import Optional, Tuple, Union

import numpy as np

from zarrita.store import Store


# ValueHandle abstracts over files, byte buffer, and arrays.
# Makes it easy pass around as references with lazy reading (esp. for files).
# Facilitates implementations of codec which either operate on arrays or bytes.
class ValueHandle:
    @abstractmethod
    def __setitem__(
        self,
        selection: Union[slice, Tuple[slice, ...]],
        value: "ValueHandle",
    ):
        pass

    @abstractmethod
    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]) -> "ValueHandle":
        pass

    @abstractmethod
    def tobytes(self) -> Optional[bytes]:
        pass

    @abstractmethod
    def toarray(self, shape: Optional[Tuple[int, ...]] = None) -> Optional[np.ndarray]:
        pass


class FileHandle(ValueHandle):
    store: "Store"
    path: str

    def __init__(self, store: "Store", path: str):
        super().__init__()
        self.store = store
        self.path = path

    def __setitem__(
        self,
        selection: Union[slice, Tuple[slice, ...]],
        value: ValueHandle,
    ):
        assert isinstance(selection, slice)
        if selection.start is None and selection.stop is None:
            buf = value.tobytes()
            if buf:
                self.store.set(self.path, buf)
            else:
                self.store.delete(self.path)
        else:
            buf = value.tobytes()
            if buf:
                self.store.set(self.path, buf, (selection.start, selection.stop))

    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]) -> ValueHandle:
        assert isinstance(selection, slice)
        buf = self.store.get(self.path, (selection.start, selection.stop))
        if buf is not None:
            return BufferHandle(buf)
        return NoneHandle()

    def tobytes(self) -> Optional[bytes]:
        return self.store.get(self.path)

    def toarray(self, shape: Optional[Tuple[int, ...]] = None) -> Optional[np.ndarray]:
        buf = self.tobytes()
        if buf is not None:
            return np.frombuffer(buf)
        return None


class BufferHandle(ValueHandle):
    buf: bytes

    def __init__(self, buf: bytes) -> None:
        super().__init__()
        self.buf = buf

    def __setitem__(
        self,
        selection: Union[slice, Tuple[slice, ...]],
        value: ValueHandle,
    ):
        assert isinstance(selection, slice)
        buf = value.tobytes()
        if buf:
            if selection.start is None and selection.stop is None:
                self.buf = buf
            else:
                tmp = bytearray(self.buf)
                tmp[selection] = buf
                self.buf = bytes(tmp)

    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]) -> ValueHandle:
        assert isinstance(selection, slice)
        return BufferHandle(self.buf[selection])

    def tobytes(self) -> Optional[bytes]:
        return self.buf

    def toarray(self, shape: Optional[Tuple[int, ...]] = None) -> Optional[np.ndarray]:
        return np.frombuffer(self.buf)


class ArrayHandle(ValueHandle):
    array: np.ndarray

    def __init__(self, array: np.ndarray) -> None:
        super().__init__()
        self.array = array

    def __setitem__(
        self,
        selection: Union[slice, Tuple[slice, ...]],
        value: ValueHandle,
    ):
        self.array[selection] = value.toarray()

    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]) -> ValueHandle:
        return ArrayHandle(self.array[selection])

    def tobytes(self) -> Optional[bytes]:
        array = self.array
        if not array.flags.c_contiguous and not array.flags.f_contiguous:
            array = array.copy(order="K")
        return array.tobytes()

    def toarray(self, shape: Optional[Tuple[int, ...]] = None) -> Optional[np.ndarray]:
        return self.array


class NoneHandle(ValueHandle):
    def __setitem__(
        self, selection: Union[slice, Tuple[slice, ...]], value: ValueHandle
    ):
        raise NotImplemented

    def __getitem__(self, selection: Union[slice, Tuple[slice, ...]]) -> ValueHandle:
        return self

    def tobytes(self) -> Optional[bytes]:
        return None

    def toarray(self, shape: Optional[Tuple[int, ...]] = None) -> Optional[np.ndarray]:
        return None



import json
from typing import Any, Dict, Literal, Optional, Union

from attr import asdict, field, frozen

from zarrita.array import Array
from zarrita.common import ZARR_JSON, make_cattr
from zarrita.store import Store


@frozen
class GroupMetadata:
    attributes: Dict[str, Any] = field(factory=dict)
    zarr_format: Literal[3] = 3
    node_type: Literal["group"] = "group"


class Group:
    metadata: GroupMetadata
    store: Store
    path: str

    @classmethod
    def create(
        cls, store: "Store", path: str, *, attributes: Optional[Dict[str, Any]] = None
    ) -> "Group":
        group = cls()
        group.metadata = GroupMetadata(attributes=attributes or {})
        group.store = store
        group.path = path
        group._save_metadata()
        return group

    @classmethod
    def open(cls, store: "Store", path: str) -> "Group":
        zarr_json_bytes = store.get(f"{path}/{ZARR_JSON}")
        assert zarr_json_bytes is not None
        return cls.from_json(store, path, json.loads(zarr_json_bytes))

    @classmethod
    def from_json(cls, store: Store, path: str, zarr_json: Any) -> "Group":
        group = cls()
        group.metadata = make_cattr().structure(zarr_json, GroupMetadata)
        group.store = store
        group.path = path
        return group

    @staticmethod
    def open_or_array(store: Store, path: str) -> Union[Array, "Group"]:
        zarr_json_bytes = store.get(f"{path}/{ZARR_JSON}")
        assert zarr_json_bytes is not None
        zarr_json = json.loads(zarr_json_bytes)
        if zarr_json["node_type"] == "group":
            return Group.from_json(store, path, zarr_json)
        if zarr_json["node_type"] == "array":
            return Array.from_json(store, path, zarr_json)
        raise KeyError

    def _save_metadata(self) -> None:
        self.store.set(
            f"{self.path}/{ZARR_JSON}",
            json.dumps(asdict(self.metadata)).encode(),
        )

    def _dereference_path(self, path: str) -> str:
        assert isinstance(path, str)
        path = f"{self.path}/{path}"
        if len(path) > 1:
            assert path[-1] != "/"
        return path

    def __getitem__(self, path: str) -> Union[Array, "Group"]:
        path = self._dereference_path(path)
        return Group.open_or_array(self.store, path)

    def create_group(self, path: str, **kwargs) -> "Group":
        path = self._dereference_path(path)
        return Group.create(self.store, path, **kwargs)

    def create_array(self, path: str, **kwargs) -> Array:
        path = self._dereference_path(path)
        return Array.create(self.store, path, **kwargs)

    def __repr__(self):
        path = self.path
        return f"<Group {path}>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zarrita']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.2.0,<23.0.0',
 'cattrs>=22.2.0,<23.0.0',
 'fsspec>=2022.2.0,<2023.0.0',
 'numcodecs>=0.10.0,<0.11.0',
 'numpy>=1.24.2,<2.0.0']

setup_kwargs = {
    'name': 'zarrita',
    'version': '0.1.0a1',
    'description': '',
    'long_description': "# Zarrita\n\nZarrita is an experimental implementation of [Zarr v3](https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html) including [sharding](https://zarr.dev/zeps/draft/ZEP0002.html). This is only a technical proof of concept meant for generating sample datasets. Not recommended for production use.\n\n## Setup\n\n```python\nimport zarrita\nimport numpy as np\n\nstore = zarrita.FileSystemStore('file://./testdata')\n```\n\n## Create an array\n\n```python\na = zarrita.Array.create(\n    store,\n    'array',\n    shape=(6, 10),\n    dtype='int32',\n    chunk_shape=(2, 5),\n    codecs=[zarrita.codecs.gzip_codec(level=1)],\n    attributes={'question': 'life', 'answer': 42}\n)\na[:, :] = np.ones((6, 10), dtype='int32')\n```\n\n## Open an array\n\n```python\na = zarrita.Array.open(store, 'array')\nassert np.array_equal(a[:, :], np.ones((6, 10), dtype='int32'))\n```\n\n## Create an array with sharding\n\n```python\na = zarrita.Array.create(\n    store,\n    'sharding',\n    shape=(16, 16),\n    dtype='int32',\n    chunk_shape=(16, 16),\n    chunk_key_encoding=('v2', '.'),\n    codecs=[\n        zarrita.codecs.sharding_codec(\n            chunk_shape=(8, 8),\n            codecs=[zarrita.codecs.gzip_codec(level=1)]\n        ),\n    ],\n)\ndata = np.arange(0, 16 * 16, dtype='int32').reshape((16, 16))\na[:, :] = data\nassert np.array_equal(a[:, :], data)\n```\n\n## Create a group\n\n```python\ng = zarrita.Group.create(store, 'group')\ng2 = g.create_group('group2')\na = g2.create_array(\n    'array',\n    shape=(16, 16),\n    dtype='int32',\n    chunk_shape=(16, 16),\n)\na[:, :] = np.arange(0, 16 * 16, dtype='int32').reshape((16, 16))\n```\n\n## Open a group\n\n```python\ng = zarrita.Group.open(store, 'group')\ng2 = g['group2']\na = g['group2/array']\nassert np.array_equal(a[:, :], np.arange(0, 16 * 16, dtype='int32').reshape((16, 16)))\n```\n\n# Credits\n\nThis is a largely-rewritten fork of `zarrita` by [@alimanfoo](https://github.com/alimanfoo). It implements the Zarr v3 draft specification created by [@alimanfoo](https://github.com/alimanfoo), [@jstriebel](https://github.com/jstriebel), [@jbms](https://github.com/jbms) et al.\n\nLicensed under MIT\n",
    'author': 'Norman Rzepka',
    'author_email': 'code@normanrz.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import itertools
import math
import numbers
from typing import Any, NamedTuple, Tuple, Union


def _ensure_tuple(v):
    if not isinstance(v, tuple):
        v = (v,)
    return v


def _err_too_many_indices(selection, shape):
    raise IndexError(
        "too many indices for array; expected {}, got {}".format(
            len(shape), len(selection)
        )
    )


def _err_boundscheck(dim_len):
    raise IndexError("index out of bounds for dimension with length {}".format(dim_len))


def _err_negative_step():
    raise IndexError("only slices with step >= 1 are supported")


def _check_selection_length(selection, shape):
    if len(selection) > len(shape):
        _err_too_many_indices(selection, shape)


def _replace_ellipsis(selection, shape) -> Tuple[int, ...]:
    selection = _ensure_tuple(selection)

    # count number of ellipsis present
    n_ellipsis = sum(1 for i in selection if i is Ellipsis)

    if n_ellipsis > 1:
        # more than 1 is an error
        raise IndexError("an index can only have a single ellipsis ('...')")

    elif n_ellipsis == 1:
        # locate the ellipsis, count how many items to left and right
        n_items_l = selection.index(Ellipsis)  # items to left of ellipsis
        n_items_r = len(selection) - (n_items_l + 1)  # items to right of ellipsis
        n_items = len(selection) - 1  # all non-ellipsis items

        if n_items >= len(shape):
            # ellipsis does nothing, just remove it
            selection = tuple(i for i in selection if i != Ellipsis)

        else:
            # replace ellipsis with as many slices are needed for number of dims
            new_item = selection[:n_items_l] + ((slice(None),) * (len(shape) - n_items))
            if n_items_r:
                new_item += selection[-n_items_r:]
            selection = new_item

    # fill out selection if not completely specified
    if len(selection) < len(shape):
        selection += (slice(None),) * (len(shape) - len(selection))

    # check selection not too long
    _check_selection_length(selection, shape)

    return selection


class _ChunkDimProjection(NamedTuple):
    dim_chunk_ix: Any
    dim_chunk_sel: Any
    dim_out_sel: Any


def _normalize_integer_selection(dim_sel, dim_len):
    # normalize type to int
    dim_sel = int(dim_sel)

    # handle wraparound
    if dim_sel < 0:
        dim_sel = dim_len + dim_sel

    # handle out of bounds
    if dim_sel >= dim_len or dim_sel < 0:
        _err_boundscheck(dim_len)

    return dim_sel


class _IntDimIndexer:
    def __init__(self, dim_sel, dim_len, dim_chunk_len):
        # normalize
        dim_sel = _normalize_integer_selection(dim_sel, dim_len)

        # store attributes
        self.dim_sel = dim_sel
        self.dim_len = dim_len
        self.dim_chunk_len = dim_chunk_len
        self.nitems = 1

    def __iter__(self):
        dim_chunk_ix = self.dim_sel // self.dim_chunk_len
        dim_offset = dim_chunk_ix * self.dim_chunk_len
        dim_chunk_sel = self.dim_sel - dim_offset
        dim_out_sel = None
        yield _ChunkDimProjection(dim_chunk_ix, dim_chunk_sel, dim_out_sel)


def _ceildiv(a, b):
    return math.ceil(a / b)


class _SliceDimIndexer:
    def __init__(self, dim_sel, dim_len, dim_chunk_len):
        # normalize
        self.start, self.stop, self.step = dim_sel.indices(dim_len)
        if self.step < 1:
            _err_negative_step()

        # store attributes
        self.dim_len = dim_len
        self.dim_chunk_len = dim_chunk_len
        self.nitems = max(0, _ceildiv((self.stop - self.start), self.step))
        self.nchunks = _ceildiv(self.dim_len, self.dim_chunk_len)

    def __iter__(self):
        # figure out the range of chunks we need to visit
        dim_chunk_ix_from = self.start // self.dim_chunk_len
        dim_chunk_ix_to = _ceildiv(self.stop, self.dim_chunk_len)

        # iterate over chunks in range
        for dim_chunk_ix in range(dim_chunk_ix_from, dim_chunk_ix_to):
            # compute offsets for chunk within overall array
            dim_offset = dim_chunk_ix * self.dim_chunk_len
            dim_limit = min(self.dim_len, (dim_chunk_ix + 1) * self.dim_chunk_len)

            # determine chunk length, accounting for trailing chunk
            dim_chunk_len = dim_limit - dim_offset

            if self.start < dim_offset:
                # selection starts before current chunk
                dim_chunk_sel_start = 0
                remainder = (dim_offset - self.start) % self.step
                if remainder:
                    dim_chunk_sel_start += self.step - remainder
                # compute number of previous items, provides offset into output array
                dim_out_offset = _ceildiv((dim_offset - self.start), self.step)

            else:
                # selection starts within current chunk
                dim_chunk_sel_start = self.start - dim_offset
                dim_out_offset = 0

            if self.stop > dim_limit:
                # selection ends after current chunk
                dim_chunk_sel_stop = dim_chunk_len

            else:
                # selection ends within current chunk
                dim_chunk_sel_stop = self.stop - dim_offset

            dim_chunk_sel = slice(dim_chunk_sel_start, dim_chunk_sel_stop, self.step)
            dim_chunk_nitems = _ceildiv(
                (dim_chunk_sel_stop - dim_chunk_sel_start), self.step
            )
            dim_out_sel = slice(dim_out_offset, dim_out_offset + dim_chunk_nitems)

            yield _ChunkDimProjection(dim_chunk_ix, dim_chunk_sel, dim_out_sel)


class _ChunkProjection(NamedTuple):
    chunk_coords: Any
    chunk_selection: Any
    out_selection: Any


class BasicIndexer:
    def __init__(
        self,
        selection: Union[slice, Tuple[slice, ...]],
        shape: Tuple[int, ...],
        chunk_shape: Tuple[int, ...],
    ):
        # handle ellipsis
        selection2 = _replace_ellipsis(selection, shape)

        # setup per-dimension indexers
        dim_indexers = []
        for dim_sel, dim_len, dim_chunk_len in zip(selection2, shape, chunk_shape):
            if isinstance(dim_sel, numbers.Integral):
                dim_indexer = _IntDimIndexer(dim_sel, dim_len, dim_chunk_len)

            elif isinstance(dim_sel, slice):
                dim_indexer = _SliceDimIndexer(dim_sel, dim_len, dim_chunk_len)

            else:
                raise IndexError(
                    "unsupported selection item for basic indexing; "
                    "expected integer or slice, got {!r}".format(type(dim_sel))
                )

            dim_indexers.append(dim_indexer)

        self.dim_indexers = dim_indexers
        self.shape = tuple(
            s.nitems for s in self.dim_indexers if not isinstance(s, _IntDimIndexer)
        )

    def __iter__(self):
        for dim_projections in itertools.product(*self.dim_indexers):
            chunk_coords = tuple(p.dim_chunk_ix for p in dim_projections)
            chunk_selection = tuple(p.dim_chunk_sel for p in dim_projections)
            out_selection = tuple(
                p.dim_out_sel for p in dim_projections if p.dim_out_sel is not None
            )

            yield _ChunkProjection(chunk_coords, chunk_selection, out_selection)
